import copy
from typing import Callable, Generic, List, Optional, Tuple, TypeVar

from caplena.constants import LIST_PAGINATION_LIMIT

T = TypeVar("T")


class CaplenaIterator(Generic[T]):
    """A lazy iterator, only fetches more results when the entries are iterated."""

    @property
    def count(self) -> int:
        """The total number of elements that exist for the requested resource."""
        if self._total_count is not None:
            return self._total_count
        else:
            self._retrieve_next_page()
            return self._total_count  # type: ignore

    def __init__(
        self,
        *,
        results_fetcher: Callable[[int], Tuple[List[T], bool, int]],
        current_page: int = 0,
        total_results_fetched: int = 0,
        results: Optional[List[T]] = None,
        total_count: Optional[int] = None,
        limit: Optional[int] = None,
        has_next: bool = True,
    ):
        self._results_fetcher = results_fetcher
        self._limit = limit

        self._total_results_iterated = 0
        self._current_results_index = 0
        self._default_pagination_limit = LIST_PAGINATION_LIMIT

        self._current_page = current_page
        self._total_results_fetched = total_results_fetched
        self._results: List[T] = [] if results is None else results
        self._total_count: Optional[int] = total_count
        self._has_next: bool = has_next

    def __str__(self) -> str:
        # note: if nothing has been fetched yet, this will trigger the inital fetch
        count = self.count

        results = ", ".join([str(res) for res in self._results])
        if len(self._results) < count:
            results += ", ..."
        return f"Iterator(count={count}, results=[{results}])"

    def __repr__(self) -> str:
        return self.__str__()

    def _retrieve_next_page(self) -> None:
        self._current_page += 1
        results, has_next, count = self._results_fetcher(self._current_page)

        self._total_results_fetched += len(results)
        self._current_results_index = 0
        self._results = results
        self._total_count = count
        self._has_next = has_next

    def __len__(self) -> int:
        return self.count if self._limit is None or self.count < self._limit else self._limit

    def __iter__(self) -> "CaplenaIterator[T]":
        return type(self)(
            results_fetcher=self._results_fetcher,
            limit=self._limit,
            current_page=self._current_page,
            total_results_fetched=self._total_results_fetched,
            results=copy.deepcopy(self._results),
            total_count=self._total_count,
            has_next=self._has_next,
        )

    def __next__(self) -> T:
        if self._limit and self._total_results_iterated >= self._limit:
            raise StopIteration()

        self._total_results_iterated += 1
        if self._total_results_iterated > self._total_results_fetched and self._has_next:
            self._retrieve_next_page()
        elif self._total_results_iterated > self._total_results_fetched:
            raise StopIteration()

        self._current_results_index += 1
        return self._results[self._current_results_index - 1]

    def __getitem__(self, key: int) -> T:
        pagination_limit = (
            self._limit if self._limit is not None else self._default_pagination_limit
        )
        target_page = key // pagination_limit + 1
        item_index = key - (target_page * pagination_limit)
        if target_page == self._current_page and len(self._results) > 0:
            return self._results[item_index]
        else:
            results, has_next, count = self._results_fetcher(target_page)
            return results[item_index]



from typing import Type, Union

from caplena.api import ApiBaseUri, ApiRequestor, ApiVersion
from caplena.http.http_client import HttpClient, HttpRetry
from caplena.logging.default_logger import DefaultLogger
from caplena.logging.logger import Logger, LoggingLevel


class Configuration:
    @property
    def api_key(self) -> str:
        return self._api_key

    @property
    def api_base_uri(self) -> ApiBaseUri:
        return self._api_base_uri

    @property
    def api_version(self) -> ApiVersion:
        return self._api_version

    @property
    def timeout(self) -> int:
        return self._timeout

    @property
    def max_retries(self) -> int:
        return self._max_retries

    @property
    def backoff_factor(self) -> float:
        return self._backoff_factor

    @property
    def http_client(self) -> HttpClient:
        return self._http_client

    @property
    def api_requestor(self) -> ApiRequestor:
        return self._api_requestor

    @property
    def logging_level(self) -> LoggingLevel:
        return self._logging_level

    @property
    def logger(self) -> Logger:
        return self._logger

    def __init__(
        self,
        *,
        api_key: str,
        http_client: Union[Type[HttpClient], HttpClient],
        api_base_uri: ApiBaseUri = ApiBaseUri.PRODUCTION,
        api_version: ApiVersion = ApiVersion.VER_2022_06_20,
        timeout: int = HttpClient.DEFAULT_TIMEOUT,
        max_retries: int = HttpRetry.DEFAULT_MAX_RETRIES,
        backoff_factor: float = HttpRetry.DEFAULT_BACKOFF_FACTOR,
        logging_level: LoggingLevel = LoggingLevel.WARNING,
    ):
        self._api_key = api_key
        self._api_base_uri = api_base_uri
        self._api_version = api_version
        self._timeout = timeout
        self._max_retries = max_retries
        self._backoff_factor = backoff_factor
        self._logging_level = logging_level

        self._logger = DefaultLogger("caplena", self._logging_level)
        self._http_client = self.build_http_client(
            http_client,
            logger=self._logger,
            timeout=timeout,
            max_retries=max_retries,
            backoff_factor=backoff_factor,
        )
        self._api_requestor = ApiRequestor(
            http_client=self._http_client,
            logger=self._logger,
        )

    @staticmethod
    def build_http_client(
        http_client: Union[Type[HttpClient], HttpClient],
        *,
        logger: Logger,
        timeout: int = HttpClient.DEFAULT_TIMEOUT,
        max_retries: int = HttpRetry.DEFAULT_MAX_RETRIES,
        backoff_factor: float = HttpRetry.DEFAULT_BACKOFF_FACTOR,
    ) -> HttpClient:
        # check if we get http client instance or if we should instantiate it ourselves
        if not isinstance(http_client, HttpClient):
            http_client = http_client()

        http_client.logger = logger
        http_client.timeout = timeout
        http_client.retry.max_retries = max_retries
        http_client.retry.backoff_factor = backoff_factor

        return http_client



from typing import Any, Dict, List, Optional, Union

from caplena.api.api_base_uri import ApiBaseUri
from caplena.api.api_exception import ApiException
from caplena.api.api_filter import ApiFilter
from caplena.api.api_ordering import ApiOrdering
from caplena.api.api_version import ApiVersion
from caplena.helpers import Helpers
from caplena.http.http_client import HttpClient, HttpMethod, HttpRetry
from caplena.http.http_response import HttpResponse
from caplena.logging.logger import Logger


class ApiRequestor:
    def __init__(
        self,
        *,
        http_client: HttpClient,
        logger: Logger,
    ):
        self.http_client = http_client
        self.logger = logger

    def build_payload(self, **kwargs: Any) -> Dict[str, Any]:
        return Helpers.build_dict(**kwargs)

    def build_uri(
        self,
        *,
        base_uri: Union[str, ApiBaseUri],
        path: str,
        path_params: Optional[Dict[str, str]] = None,
        query_params: Optional[Dict[str, str]] = None,
    ) -> str:
        path_params = path_params if path_params is not None else {}
        query_params = query_params if query_params is not None else {}
        if isinstance(base_uri, ApiBaseUri):
            base_uri = base_uri.url

        absolute_uri = Helpers.append_path(base_uri=base_uri, path=path)
        return Helpers.build_qualified_uri(
            absolute_uri,
            path_params=path_params,
            query_params=query_params,
        )

    def build_query_params(
        self,
        *,
        query_params: Optional[Dict[str, str]] = None,
        filter: Optional[ApiFilter] = None,
        order_by: Optional[ApiOrdering] = None,
    ) -> Dict[str, str]:
        new_query_params: Dict[str, str] = {}

        if filter is not None:
            new_query_params.update(filter.to_query_params())
        if order_by is not None:
            new_query_params.update(order_by.to_query_params())
        if query_params is not None:
            new_query_params.update(query_params)

        return new_query_params

    def build_request_headers(
        self,
        *,
        headers: Optional[Dict[str, str]] = None,
        api_version: ApiVersion = ApiVersion.DEFAULT,
        api_key: Optional[str] = None,
    ) -> Dict[str, str]:
        headers = {} if headers is None else headers.copy()
        headers.setdefault("Accept", "application/json")

        # note: we do not allow clients to overwrite user-agent, caplena-api-key or caplena-api-version
        headers["User-Agent"] = Helpers.get_user_agent(identifier=self.http_client.identifier)
        if api_key is not None:
            headers["Caplena-API-Key"] = api_key
        if api_version != ApiVersion.DEFAULT:
            headers["Caplena-API-Version"] = api_version.version

        return headers

    def build_exc(self, response: HttpResponse) -> ApiException:
        exc_body = response.json
        if exc_body:
            self.logger.info(
                "Received error from server", type=exc_body["type"], code=exc_body["code"]
            )
            return ApiException(
                type=exc_body["type"],
                code=exc_body["code"],
                message=exc_body["message"],
                details=exc_body.get("details"),
                help=exc_body.get("help"),
                context=exc_body.get("context"),
            )
        else:
            return ApiException(type="internal_error", code="body.invalid_format")

    def request_raw(
        self,
        base_uri: Union[str, ApiBaseUri],
        path: str,
        *,
        method: HttpMethod = HttpMethod.GET,
        api_version: ApiVersion = ApiVersion.DEFAULT,
        api_key: Optional[str] = None,
        path_params: Optional[Dict[str, str]] = None,
        query_params: Optional[Dict[str, str]] = None,
        json: Optional[Union[Dict[str, Any], List[Any]]] = None,
        headers: Optional[Dict[str, str]] = None,
        timeout: Optional[int] = None,
        retry: Optional[HttpRetry] = None,
    ) -> HttpResponse:
        absolute_uri = self.build_uri(
            base_uri=base_uri,
            path=path,
            path_params=path_params,
            query_params=query_params,
        )
        headers = self.build_request_headers(
            headers=headers,
            api_version=api_version,
            api_key=api_key,
        )

        return self.http_client.request(
            uri=absolute_uri,
            method=method,
            headers=headers,
            json=json,
            timeout=timeout,
            retry=retry,
        )

    def get(
        self,
        base_uri: Union[str, ApiBaseUri],
        path: str,
        *,
        api_version: ApiVersion = ApiVersion.DEFAULT,
        api_key: Optional[str] = None,
        path_params: Optional[Dict[str, str]] = None,
        query_params: Optional[Dict[str, str]] = None,
        headers: Optional[Dict[str, str]] = None,
        filter: Optional[ApiFilter] = None,
        order_by: Optional[ApiOrdering] = None,
        timeout: Optional[int] = None,
        retry: Optional[HttpRetry] = None,
    ) -> HttpResponse:
        query_params = self.build_query_params(
            filter=filter, order_by=order_by, query_params=query_params
        )
        return self.request_raw(
            base_uri=base_uri,
            path=path,
            method=HttpMethod.GET,
            api_version=api_version,
            api_key=api_key,
            path_params=path_params,
            query_params=query_params,
            json=None,
            headers=headers,
            timeout=timeout,
            retry=retry,
        )

    def post(
        self,
        base_uri: Union[str, ApiBaseUri],
        path: str,
        *,
        api_version: ApiVersion = ApiVersion.DEFAULT,
        api_key: Optional[str] = None,
        path_params: Optional[Dict[str, str]] = None,
        query_params: Optional[Dict[str, str]] = None,
        json: Optional[Union[Dict[str, Any], List[Any]]] = None,
        headers: Optional[Dict[str, str]] = None,
        timeout: Optional[int] = None,
        retry: Optional[HttpRetry] = None,
    ) -> HttpResponse:
        return self.request_raw(
            base_uri=base_uri,
            path=path,
            method=HttpMethod.POST,
            api_version=api_version,
            api_key=api_key,
            path_params=path_params,
            query_params=query_params,
            json=json,
            headers=headers,
            timeout=timeout,
            retry=retry,
        )

    def put(
        self,
        base_uri: Union[str, ApiBaseUri],
        path: str,
        *,
        api_version: ApiVersion = ApiVersion.DEFAULT,
        api_key: Optional[str] = None,
        path_params: Optional[Dict[str, str]] = None,
        query_params: Optional[Dict[str, str]] = None,
        json: Optional[Union[Dict[str, Any], List[Any]]] = None,
        headers: Optional[Dict[str, str]] = None,
        timeout: Optional[int] = None,
        retry: Optional[HttpRetry] = None,
    ) -> HttpResponse:
        return self.request_raw(
            base_uri=base_uri,
            path=path,
            method=HttpMethod.PUT,
            api_version=api_version,
            api_key=api_key,
            path_params=path_params,
            query_params=query_params,
            json=json,
            headers=headers,
            timeout=timeout,
            retry=retry,
        )

    def patch(
        self,
        base_uri: Union[str, ApiBaseUri],
        path: str,
        *,
        api_version: ApiVersion = ApiVersion.DEFAULT,
        api_key: Optional[str] = None,
        path_params: Optional[Dict[str, str]] = None,
        query_params: Optional[Dict[str, str]] = None,
        json: Optional[Union[Dict[str, Any], List[Any]]] = None,
        headers: Optional[Dict[str, str]] = None,
        timeout: Optional[int] = None,
        retry: Optional[HttpRetry] = None,
    ) -> HttpResponse:
        return self.request_raw(
            base_uri=base_uri,
            path=path,
            method=HttpMethod.PATCH,
            api_version=api_version,
            api_key=api_key,
            path_params=path_params,
            query_params=query_params,
            json=json,
            headers=headers,
            timeout=timeout,
            retry=retry,
        )

    def delete(
        self,
        base_uri: Union[str, ApiBaseUri],
        path: str,
        *,
        api_version: ApiVersion = ApiVersion.DEFAULT,
        api_key: Optional[str] = None,
        path_params: Optional[Dict[str, str]] = None,
        query_params: Optional[Dict[str, str]] = None,
        headers: Optional[Dict[str, str]] = None,
        timeout: Optional[int] = None,
        retry: Optional[HttpRetry] = None,
    ) -> HttpResponse:
        return self.request_raw(
            base_uri=base_uri,
            path=path,
            method=HttpMethod.DELETE,
            api_version=api_version,
            api_key=api_key,
            path_params=path_params,
            query_params=query_params,
            json=None,
            headers=headers,
            timeout=timeout,
            retry=retry,
        )



"""Python Library for Caplena REST API."""

from caplena.api import ApiBaseUri, ApiOrdering, ApiVersion
from caplena.client import Client
from caplena.http.http_client import HttpMethod, HttpRetry
from caplena.http.requests_http_client import RequestsHttpClient
from caplena.logging.logger import LoggingLevel
from caplena.version import __version__

__all__ = [
    "__version__",
    "Client",
    "ApiBaseUri",
    "ApiVersion",
    "ApiOrdering",
    "HttpRetry",
    "HttpMethod",
    "RequestsHttpClient",
    "LoggingLevel",
]



import json
from typing import Any, ClassVar, Optional


class ApiException(Exception):
    """The API Exception that is thrown by the Caplena REST API.

    :param type: The type of the error returned.
    :param code: A short string that uniquely identifies the reason for this error.
        Useful for programmatic error handling. Should never be shown to your end users.
    :param message: A brief human-readable message providing more details about the error
        that has occurred. Please note that error messages might change and are therefore
        not suitable for programmatic error handling. Can be shown to end users.
    :param details: A lengthier human-readable explanation of the error. This property
        is intened for use by developers only and provides additional information on how
        this issue could be resolved. Should never be shown to your end users.
    :param help: URL that links to part of our developer documentation, allowing engineers
        to learn more about how to fix the given issue. Should never be shown
        to your end users.
    :param context: Additional context that might be present depending on the error type and code.
    """

    DEFAULT_MESSAGE: ClassVar[
        str
    ] = "An unknown error occurred. Please reach out to us at support@caplena.com."

    type: str
    """The type of the error returned."""

    code: str
    """A short string that uniquely identifies the reason for this error.
    Useful for programmatic error handling. Should never be shown to your end users.
    """

    message: str
    """A brief human-readable message providing more details about the error
    that has occurred. Please note that error messages might change and are therefore
    not suitable for programmatic error handling. Can be shown to end users.
    """

    details: Optional[str]
    """A lengthier human-readable explanation of the error. This property
    is intened for use by developers only and provides additional information on how
    this issue could be resolved. Should never be shown to your end users.
    """

    help: Optional[str]
    """URL that links to part of our developer documentation, allowing engineers
    to learn more about how to fix the given issue. Should never be shown
    to your end users.
    """

    context: Any
    """Additional context that might be present depending on the error type and code."""

    def __init__(
        self,
        *,
        type: str,
        code: str,
        message: str = DEFAULT_MESSAGE,
        details: Optional[str] = None,
        help: Optional[str] = None,
        context: Any = None,
    ):
        msg = f"{type}[{code}]: {message}"
        if details:
            msg += " " + details
        if help:
            msg += f" For more information, please visit {help}."
        if context:
            msg += f" (context={json.dumps(context)})"
        super().__init__(msg)

        self.type = type
        self.code = code
        self.message = message
        self.details = details
        self.help = help
        self.context = context



import platform
import re
import sys
from datetime import datetime
from typing import Any, Dict, Iterable
from urllib.parse import urlencode

from caplena.constants import NOT_SET
from caplena.version import __version__


class Helpers:
    @staticmethod
    def get_user_agent(identifier: str) -> str:
        client_info = f"{identifier}/{__version__}"
        python_info = "python/{ver.major}.{ver.minor}.{ver.micro}".format(ver=sys.version_info)
        system_info = f"{platform.system()}/{platform.release()}"
        return " ".join([client_info, python_info, system_info])

    @staticmethod
    def append_path(base_uri: str, path: str) -> str:
        absolute_uri = base_uri
        if path:
            base_uri = base_uri if base_uri[-1] != "/" else base_uri[:-1]
            path = path if path[0] != "/" else path[1:]
            absolute_uri = f"{base_uri}/{path}"

        return absolute_uri

    @staticmethod
    def partial_dict(dict: Dict[str, Any], attrs: Iterable[str]) -> Dict[str, Any]:
        partial: Dict[str, Any] = {}
        for attr in attrs:
            partial[attr] = dict[attr]
        return partial

    @staticmethod
    def from_rfc3339_datetime(value: str) -> datetime:
        iso_8601 = value.replace("Z", "+00:00")
        return datetime.fromisoformat(iso_8601)

    @staticmethod
    def to_rfc3339_datetime(dt: datetime) -> str:
        rfc3339 = dt.strftime("%Y-%m-%dT%H:%M:%S.")
        rfc3339 += dt.strftime("%f")[:3]

        tz = dt.strftime("%z")
        if tz == "+0000":
            tz = "Z"
        elif tz != "":
            tz = tz[:3] + ":" + tz[3:]
        return rfc3339 + tz

    @staticmethod
    def build_qualified_uri(
        uri: str, *, path_params: Dict[str, str], query_params: Dict[str, str]
    ) -> str:
        # constructing path parameters
        for param_key, param_value in path_params.items():
            uri = uri.replace("{" + param_key + "}", str(param_value))
        has_remaining_path_params = re.search(r"{(.*?)}", uri)
        if has_remaining_path_params:
            found_param = has_remaining_path_params.group(1)
            if found_param == "":
                raise ValueError(
                    "Path specifies an invalid path parameter. Parameter name must not be empty."
                )
            else:
                raise ValueError(
                    f"Path requires `{has_remaining_path_params.group(1)}` parameter, but none is given."
                )

        # constructing query parameters
        query_string = urlencode(query_params)
        if query_string:
            uri += "?" + query_string

        return uri

    @staticmethod
    def build_escaped_filter_str(value: str) -> str:
        escaped = value.replace("\\", "\\\\")
        return re.sub(r"(:|,|;)", r"\\\1", escaped)

    @staticmethod
    def build_dict(**kwargs: Any) -> Dict[str, Any]:
        constructed_dict: Dict[str, Any] = {}
        for field_name, field in kwargs.items():
            if field != NOT_SET:
                constructed_dict[field_name] = field
        return constructed_dict



import copy
from datetime import datetime
from typing import Any, ClassVar, Dict, List, Optional, Type, TypeVar, Union

from caplena.helpers import Helpers

T = TypeVar("T")
U = TypeVar("U", bound="ApiFilter")

ZeroOrMany = Optional[Union[T, List[T]]]
Constraints = Dict[str, List[Dict[str, List[Any]]]]


class ApiFilter:
    DEFAULT: ClassVar[str] = "__default__"

    def __init__(
        self,
        constraints: Optional[Constraints] = None,
        has_conjunction: bool = False,
    ):
        # TODO: how to restrict AND, OR, NONE operations
        self._constraints: Constraints = constraints if constraints is not None else {}
        self._has_conjunction = has_conjunction

    def to_query_params(self) -> Dict[str, str]:
        query_params: Dict[str, str] = {}
        for query_param, clauses in self._constraints.items():
            stringified_clauses: List[str] = []
            for clause in clauses:
                stringified_literals: List[str] = []
                for modifier, values in clause.items():
                    for value in values:
                        str_value = Helpers.build_escaped_filter_str(self.to_string(value=value))
                        if modifier != self.DEFAULT:
                            str_value = f"{modifier}:" + str_value
                        stringified_literals.append(str_value)
                stringified_clauses.append(",".join(stringified_literals))
            query_params[query_param] = ";".join(stringified_clauses)
        return query_params

    def __str__(self) -> str:
        stringified_clauses: List[str] = []
        for name, clauses in self._constraints.items():
            for clause in clauses:
                stringified_literals: List[str] = []
                for modifier, values in clause.items():
                    filt_name = f"{name}"
                    if modifier != self.DEFAULT:
                        filt_name += f".{modifier}"
                    str_values = ",".join([str(value) for value in values])
                    stringified_literals.append(f"{filt_name}={{{str_values}}}")
                stringified_clauses.append("(" + " | ".join(stringified_literals) + ")")
        return "ApiFilter(" + " & ".join(stringified_clauses) + ")"

    def __and__(self: U, other: U) -> U:
        """Creates and returns the immutable conjunction of two separate filters. The original
        filters will not be modified.

        :param self: The first filter.
        :type self: Filter
        :param other: The second filter.
        :type other: Filter
        :rtype: Filter
        """
        new_constraints = copy.deepcopy(self._constraints)
        other_constraints = copy.deepcopy(other._constraints)

        has_conjunction = len(new_constraints.keys()) > 0 and len(other_constraints.keys()) > 0
        has_conjunction = self._has_conjunction if self._has_conjunction else has_conjunction
        has_conjunction = other._has_conjunction if other._has_conjunction else has_conjunction

        for name, clauses in other_constraints.items():
            if name in new_constraints:
                new_constraints[name].extend(clauses)
            else:
                new_constraints[name] = clauses

        return type(self)(constraints=new_constraints, has_conjunction=has_conjunction)

    def __or__(self: U, other: U) -> U:
        """Creates and returns the immutable disjunction of two separate filters. The original
        filters will not be modified.

        :param self: The first filter.
        :type self: Filter
        :param other: The second filter.
        :type other: Filter
        :rtype: Filter
        """
        new_constraints = copy.deepcopy(self._constraints)
        other_constraints = copy.deepcopy(other._constraints)
        new_filters = list(new_constraints.keys())
        other_filters = list(other_constraints.keys())

        has_conjunction = False
        if len(new_filters) == 0:
            has_conjunction = other._has_conjunction
            new_constraints = other_constraints
        elif len(other_filters) == 0:
            has_conjunction = other._has_conjunction
        elif self._has_conjunction or other._has_conjunction:
            raise ValueError(
                "Cannot build the disjunction of already conjuncted filters To fix this error, please make sure "
                "that your filters are in conjunctive normal form (CNF)."
            )
        else:
            if len(new_filters) > 1 or len(other_filters) > 1:
                raise ValueError("Should never execute this code.")
            elif (
                len(new_filters) == 1
                and len(other_filters) == 1  # noqa: W503
                and new_filters[0] != other_filters[0]  # noqa: W503
            ):
                raise ValueError(
                    f"Cannot build the disjunction for filter `{other_filters[0]}`, as there is already a different "
                    f"filter `{new_filters[0]}` being applied."
                )
            elif len(new_filters) == 1 and len(other_filters) == 1:
                name = new_filters[0]
                for filt_name, values in other_constraints[name][0].items():
                    new_constraints[name][0].setdefault(filt_name, [])
                    new_constraints[name][0][filt_name].extend(values)

        return type(self)(constraints=new_constraints, has_conjunction=has_conjunction)

    @classmethod
    def construct(cls: Type[U], *, name: str, filters: Dict[str, ZeroOrMany[Any]]) -> U:
        constraints: Constraints = {}

        clauses: List[Dict[str, List[Any]]] = []
        for filter_name, values in filters.items():
            values_list = cls.to_list(values=values)
            if values_list is not None:
                clause: Dict[str, List[Any]] = {}
                clause[filter_name] = values_list
                clauses.append(clause)

        if len(clauses) > 0:
            constraints[name] = clauses
        has_conjunction = len(clauses) > 1
        return cls(constraints=constraints, has_conjunction=has_conjunction)

    @staticmethod
    def to_list(*, values: ZeroOrMany[Any]) -> Optional[List[Any]]:
        if values is None:
            return None
        elif isinstance(values, list) and len(values) > 0:
            return list(values)
        elif isinstance(values, list):
            return None
        else:
            return [values]

    @staticmethod
    def to_string(*, value: Any) -> str:
        if isinstance(value, datetime):
            return Helpers.to_rfc3339_datetime(value)
        else:
            return str(value)



from caplena.api.api_base_uri import ApiBaseUri
from caplena.api.api_exception import ApiException
from caplena.api.api_filter import ApiFilter, ZeroOrMany
from caplena.api.api_ordering import ApiOrdering
from caplena.api.api_requestor import ApiRequestor
from caplena.api.api_version import ApiVersion

__all__ = [
    "ApiBaseUri",
    "ApiException",
    "ApiRequestor",
    "ApiVersion",
    "ApiFilter",
    "ZeroOrMany",
    "ApiOrdering",
]



from caplena.api.api_base_uri import ApiBaseUri
from caplena.configuration import Configuration
from caplena.http.requests_http_client import RequestsHttpClient
from caplena.logging.logger import LoggingLevel

common_api_key = "4e5df2b6f642bf4f6b9eb36af587eef96dbc4b8e"

common_config = Configuration(
    api_key=common_api_key,
    http_client=RequestsHttpClient,
    api_base_uri=ApiBaseUri.LOCAL,
    logging_level=LoggingLevel.DEBUG,
)



import unittest
from datetime import datetime
from typing import Any, ClassVar, Dict, List, Optional, cast

from caplena.api.api_exception import ApiException
from caplena.controllers import ProjectsController
from caplena.filters.projects_filter import ProjectsFilter, RowsFilter
from caplena.resources import ProjectDetail, Row
from tests.common import common_config


def project_create_payload() -> Dict[str, Any]:
    return {
        "name": "Project Name",
        "tags": ["my-tag"],
        "language": "en",
        "translation_engine": None,
        "columns": [
            {
                "ref": "customer_age",
                "type": "numerical",
                "name": "Age of the customer",
            },
            {
                "ref": "our_strengths",
                "name": "Do you like us?",
                "type": "text_to_analyze",
                "topics": [
                    {
                        "label": "price",
                        "sentiment_enabled": True,
                        "category": "SERVICE",
                    },
                    {
                        "label": "network quality",
                        "sentiment_enabled": False,
                        "category": "SERVICE",
                    },
                ],
            },
            {"type": "boolean", "ref": "boolean_col", "name": "Some example boolean"},
            {"type": "text", "ref": "text_col", "name": "Some auxiliary text"},
            {"type": "date", "ref": "date_col", "name": "Some date time values."},
        ],
    }


def project_rows_create_payload() -> List[Dict[str, Any]]:
    return [
        {
            "columns": [
                {"ref": "customer_age", "value": 120},
                {"ref": "our_strengths", "value": "This is nice."},
                {"ref": "boolean_col", "value": False},
                {"ref": "text_col", "value": "iphone"},
                {"ref": "date_col", "value": datetime(year=2020, month=10, day=10, hour=17)},
            ]
        },
        {
            "columns": [
                {"ref": "customer_age", "value": None},
                {"ref": "our_strengths", "value": "Some other text."},
                {"ref": "boolean_col", "value": True},
                {"ref": "text_col", "value": "oneplus"},
                {
                    "ref": "date_col",
                    "value": datetime(year=2022, month=3, day=31, hour=14, minute=14, second=14),
                },
            ]
        },
    ]


class ProjectsControllerTests(unittest.TestCase):
    controller: ClassVar[ProjectsController]

    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()

        cls.controller = ProjectsController(config=common_config)

    def setUp(self) -> None:
        self.created_projects: List[str] = []

    def tearDown(self) -> None:
        for project_id in self.created_projects:
            try:
                self.controller.remove(id=project_id)
            except ApiException:
                print("Could not remove project with id:", project_id)

    def create_project(self, *, payload: Optional[Dict[str, Any]] = None) -> ProjectDetail:
        if payload is None:
            payload = project_create_payload()

        project = self.controller.create(**payload)
        self.created_projects.append(project.id)

        return project

    # ---- General Functionality ---- #

    def test_creating_a_project_succeeds(self) -> None:
        project = self.create_project()

        self.assertIsInstance(project.id, str)
        self.assertEqual("Project Name", project.name)
        self.assertIsInstance(project.owner, str)
        self.assertListEqual(["my-tag"], project.tags)
        self.assertEqual("pending", project.upload_status)
        self.assertEqual("en", project.language)
        self.assertEqual(None, project.translation_status)
        self.assertEqual(None, project.translation_engine)
        self.assertIsInstance(project.created, datetime)
        self.assertIsInstance(project.last_modified, datetime)

        self.assertEqual(5, len(project.columns))
        self.assertIsInstance(project.columns[0], ProjectDetail.TextToAnalyze)
        self.assertIsInstance(project.columns[1], ProjectDetail.Auxiliary)
        self.assertIsInstance(project.columns[2], ProjectDetail.Auxiliary)
        self.assertIsInstance(project.columns[3], ProjectDetail.Auxiliary)
        self.assertIsInstance(project.columns[4], ProjectDetail.Auxiliary)
        our_strengths = cast(ProjectDetail.TextToAnalyze, project.columns[0])
        customer_age = cast(ProjectDetail.Auxiliary, project.columns[1])
        boolean_col = cast(ProjectDetail.Auxiliary, project.columns[2])
        text_col = cast(ProjectDetail.Auxiliary, project.columns[3])
        date_col = cast(ProjectDetail.Auxiliary, project.columns[4])

        self.assertEqual("our_strengths", our_strengths.ref)
        self.assertEqual("Do you like us?", our_strengths.name)
        self.assertEqual("text_to_analyze", our_strengths.type)
        self.assertEqual("", our_strengths.description)
        self.assertDictEqual(
            {
                "reviewed_count": 0,
                "learns_from": None,
            },
            our_strengths.metadata.dict(),
        )
        self.assertEqual(2, len(our_strengths.topics))

        topic1, topic2 = our_strengths.topics[0], our_strengths.topics[1]
        self.assertRegex(topic1.id, r"^cd_")
        self.assertEqual("price", topic1.label)
        self.assertEqual("SERVICE", topic1.category)
        self.assertEqual("", topic1.color)
        self.assertEqual("", topic1.description)
        self.assertEqual(True, topic1.sentiment_enabled)
        self.assertDictEqual({"code": 0, "label": ""}, topic1.sentiment_neutral.dict())
        self.assertDictEqual({"code": 1, "label": ""}, topic1.sentiment_positive.dict())
        self.assertDictEqual({"code": 2, "label": ""}, topic1.sentiment_negative.dict())

        self.assertRegex(topic2.id, r"^cd_")
        self.assertEqual("network quality", topic2.label)
        self.assertEqual("SERVICE", topic2.category)
        self.assertEqual("", topic2.color)
        self.assertEqual("", topic2.description)
        self.assertEqual(False, topic2.sentiment_enabled)
        self.assertDictEqual({"code": 3, "label": ""}, topic2.sentiment_neutral.dict())
        self.assertDictEqual({"code": -1, "label": ""}, topic2.sentiment_negative.dict())
        self.assertDictEqual({"code": -1, "label": ""}, topic2.sentiment_positive.dict())

        self.assertEqual("customer_age", customer_age.ref)
        self.assertEqual("Age of the customer", customer_age.name)
        self.assertEqual("numerical", customer_age.type)

        self.assertEqual("boolean_col", boolean_col.ref)
        self.assertEqual("Some example boolean", boolean_col.name)
        self.assertEqual("boolean", boolean_col.type)

        self.assertEqual("text_col", text_col.ref)
        self.assertEqual("Some auxiliary text", text_col.name)
        self.assertEqual("text", text_col.type)

        self.assertEqual("date_col", date_col.ref)
        self.assertEqual("Some date time values.", date_col.name)
        self.assertEqual("date", date_col.type)

    def test_retrieving_a_project_succeeds(self) -> None:
        project = self.create_project()
        retrieved = self.controller.retrieve(id=project.id)

        self.assertDictEqual(project.dict(), retrieved.dict())

    def test_removing_a_project_succeeds(self) -> None:
        old_num_projects = self.controller.list(limit=1).count
        project = self.create_project()
        interim_num_projects = self.controller.list(limit=1).count
        self.controller.remove(id=project.id)
        new_num_projects = self.controller.list(limit=1).count

        self.assertEqual(old_num_projects, new_num_projects)
        self.assertEqual(old_num_projects + 1, interim_num_projects)

    def test_updating_a_project_succeeds(self) -> None:
        project_to_learn_from = self.create_project()
        project = self.create_project()

        # test: updating properties succeeds
        expected_dict = project.dict()
        project.name = "MY SUPER NOVEL PROJECT NAME"
        project.tags = ["new", "tags", "are", "cool"]
        our_strengths: ProjectDetail.TextToAnalyze = project.columns[0]  # type: ignore
        our_strengths.name = "Do you still like us?"
        our_strengths.description = "Please explain."
        our_strengths.metadata.learns_from = self.controller.build(
            ProjectDetail.TextToAnalyze.Metadata.LearnsForm,
            {"project": project_to_learn_from.id, "ref": project_to_learn_from.columns[0].ref},
        )
        project.columns[1].name = "COOL NAME"
        project.save()
        expected_dict["name"] = "MY SUPER NOVEL PROJECT NAME"
        expected_dict["tags"] = ["new", "tags", "are", "cool"]
        expected_dict["columns"][0].update(
            {"name": "Do you still like us?", "description": "Please explain."}
        )
        expected_dict["columns"][0]["metadata"].update(
            {
                "learns_from": {
                    "project": project_to_learn_from.id,
                    "ref": project_to_learn_from.columns[0].ref,
                }
            }
        )
        expected_dict["columns"][1]["name"] = "COOL NAME"
        self.assertDictEqual(project.dict(), expected_dict)

        # test: resetting learns_from succeeds
        our_strengths: ProjectDetail.TextToAnalyze = project.columns[0]  # type: ignore
        our_strengths.metadata.learns_from = None
        project.save()
        expected_dict["columns"][0]["metadata"]["learns_from"] = None

        # last modified is updated
        project_dict = project.dict()
        project_dict.pop("last_modified")
        expected_dict.pop("last_modified")
        self.assertDictEqual(project_dict, expected_dict)

    def test_listing_all_projects_succeeds(self) -> None:
        project = self.create_project()
        projects = self.controller.list(limit=1)

        self.assertIsInstance(projects.count, int)
        self.assertGreaterEqual(projects.count, 0)
        self.assertEqual(len(projects), 1)
        retrieved = (list(projects))[0]

        project_dict = project.dict()
        project_dict.pop("columns")
        self.assertDictEqual(project_dict, retrieved.dict())

    def test_filtering_projects_succeeds(self) -> None:
        filt = ProjectsFilter.language("tr")
        projects = self.controller.list(filter=filt)

        self.assertEqual(0, len(projects))
        self.assertEqual(0, projects.count)

    def test_appending_multiple_rows_succeeds(self) -> None:
        project = self.create_project()
        response = self.controller.append_rows(id=project.id, rows=project_rows_create_payload())

        self.assertEqual("pending", response.status)
        self.assertEqual(2, response.queued_rows_count)
        self.assertEqual(1.02, response.estimated_minutes)
        self.assertEqual(2, len(response.results))
        self.assertTrue(all([isinstance(row.id, str) for row in response.results]))

    def test_appending_single_row_succeeds(self) -> None:
        project = self.create_project()
        columns: List[Dict[str, Any]] = [
            {"ref": "customer_age", "value": None},
            {"ref": "our_strengths", "value": "Good price."},
            {"ref": "boolean_col", "value": False},
            {"ref": "text_col", "value": None},
            {"ref": "date_col", "value": datetime(year=2020, month=10, day=10, hour=17)},
        ]
        row = self.controller.append_row(id=project.id, columns=columns)

        self.assertIsInstance(row.id, str)
        self.assertIsInstance(row.created, datetime)
        self.assertIsInstance(row.last_modified, datetime)

        self.assertEqual(5, len(row.columns))
        self.assertIsInstance(row.columns[0], Row.TextToAnalyzeColumn)
        self.assertIsInstance(row.columns[1], Row.NumericalColumn)
        our_strengths = cast(Row.TextToAnalyzeColumn, row.columns[0])
        customer_age = cast(Row.NumericalColumn, row.columns[1])
        boolean_col = cast(Row.BooleanColumn, row.columns[2])
        text_col = cast(Row.TextColumn, row.columns[3])
        date_col = cast(Row.DateColumn, row.columns[4])

        self.assertEqual("our_strengths", our_strengths.ref)
        self.assertEqual("text_to_analyze", our_strengths.type)
        self.assertEqual("Good price.", our_strengths.value)
        self.assertEqual(False, our_strengths.was_reviewed)
        self.assertEqual(None, our_strengths.source_language)
        self.assertEqual(None, our_strengths.translated_value)
        self.assertEqual(1, len(our_strengths.topics))
        topic = our_strengths.topics[0]
        self.assertRegex(topic.id, r"^cd_")
        self.assertEqual(topic.label, "price")
        self.assertEqual(topic.category, "SERVICE")
        self.assertEqual(topic.code, 1)
        self.assertEqual(topic.sentiment_label, "")
        self.assertEqual(topic.sentiment, "positive")

        self.assertEqual("customer_age", customer_age.ref)
        self.assertEqual("numerical", customer_age.type)
        self.assertEqual(None, customer_age.value)

        self.assertEqual("boolean_col", boolean_col.ref)
        self.assertEqual("boolean", boolean_col.type)
        self.assertEqual(False, boolean_col.value)

        self.assertEqual("text_col", text_col.ref)
        self.assertEqual("text", text_col.type)
        self.assertEqual("", text_col.value)

        self.assertEqual("date_col", date_col.ref)
        self.assertEqual("date", date_col.type)
        self.assertEqual(datetime(year=2020, month=10, day=10, hour=17), date_col.value)

    def test_listing_all_rows_succeeds(self) -> None:
        project = self.create_project()
        row1 = self.controller.append_row(
            id=project.id,
            columns=[
                {"ref": "customer_age", "value": None},
                {"ref": "our_strengths", "value": "Some other text."},
                {"ref": "boolean_col", "value": False},
                {"ref": "text_col", "value": "iphone"},
                {"ref": "date_col", "value": datetime(year=2020, month=10, day=10, hour=17)},
            ],
        )
        row2 = self.controller.append_row(
            id=project.id,
            columns=[
                {"ref": "customer_age", "value": 12},
                {"ref": "our_strengths", "value": "This is my review. Very nice."},
                {"ref": "boolean_col", "value": True},
                {"ref": "text_col", "value": "samsung"},
                {"ref": "date_col", "value": datetime(year=2000, month=4, day=4, hour=4)},
            ],
        )
        rows = [row1.dict(), row2.dict()]
        retrieved = self.controller.list_rows(id=project.id, limit=2)

        self.assertEqual(2, retrieved.count)
        self.assertEqual(2, len(retrieved))

        retrieved_dict = [row.dict() for row in retrieved]
        self.assertListEqual(rows, retrieved_dict)

    def test_filtering_rows_succeeds(self) -> None:
        project = self.create_project()
        self.controller.append_row(
            id=project.id,
            columns=[
                {"ref": "customer_age", "value": None},
                {"ref": "our_strengths", "value": "Some other text."},
                {"ref": "boolean_col", "value": False},
                {"ref": "text_col", "value": "iphone"},
                {"ref": "date_col", "value": datetime(year=2020, month=10, day=10, hour=17)},
            ],
        )

        with_results = self.controller.list_rows(id=project.id)
        self.assertEqual(1, with_results.count)

        filt = RowsFilter.Columns.text_to_analyze(ref="our_strengths", contains__i=["some"])
        filtered_results = self.controller.list_rows(id=project.id, filter=filt)
        self.assertEqual(1, filtered_results.count)

        filt = RowsFilter.Columns.text_to_analyze(ref="our_strengths", was_reviewed=True)
        no_results = self.controller.list_rows(id=project.id, filter=filt)
        self.assertEqual(0, no_results.count)

    def test_retrieving_a_row_succeeds(self) -> None:
        project = self.create_project()
        row = self.controller.append_row(
            id=project.id,
            columns=[
                {"ref": "customer_age", "value": 400},
                {"ref": "our_strengths", "value": "Some other text."},
                {"ref": "boolean_col", "value": False},
                {"ref": "text_col", "value": "iphone"},
                {"ref": "date_col", "value": datetime(year=2020, month=10, day=10, hour=17)},
            ],
        )
        retrieved = self.controller.retrieve_row(p_id=project.id, r_id=row.id)

        self.assertDictEqual(row.dict(), retrieved.dict())

    def test_removing_a_row_succeeds(self) -> None:
        project = self.create_project()

        old_num_rows = self.controller.list_rows(id=project.id, limit=1).count
        row = self.controller.append_row(
            id=project.id,
            columns=[
                {"ref": "customer_age", "value": 400},
                {"ref": "our_strengths", "value": "Some other text."},
                {"ref": "boolean_col", "value": False},
                {"ref": "text_col", "value": "iphone"},
                {"ref": "date_col", "value": datetime(year=2020, month=10, day=10, hour=17)},
            ],
        )
        interim_num_rows = self.controller.list_rows(id=project.id, limit=1).count
        self.controller.remove_row(p_id=project.id, r_id=row.id)
        new_num_rows = self.controller.list_rows(id=project.id, limit=1).count

        self.assertEqual(old_num_rows, new_num_rows)
        self.assertEqual(old_num_rows + 1, interim_num_rows)

    def test_updating_a_row_succeeds(self) -> None:
        project = self.create_project()
        row = self.controller.append_row(
            id=project.id,
            columns=[
                {"ref": "customer_age", "value": 400},
                {"ref": "our_strengths", "value": "Some other text."},
                {"ref": "boolean_col", "value": False},
                {"ref": "text_col", "value": "iphone"},
                {"ref": "date_col", "value": datetime(year=2020, month=10, day=10, hour=17)},
            ],
        )
        expected_dict = row.dict()
        our_strengths: Row.TextToAnalyzeColumn = row.columns[0]  # type: ignore
        our_strengths.value = "this is a new text value."
        our_strengths.was_reviewed = True
        customer_age: Row.NumericalColumn = row.columns[1]  # type: ignore
        customer_age.value = 100000

        row.save()
        row_dict = row.dict()
        expected_dict["columns"][0].update(
            {"value": "this is a new text value.", "was_reviewed": True}
        )
        computed_row_fields = {"last_modified"}
        # computed fields are updated when the value is changed, so don't compare them
        for field in computed_row_fields:
            expected_dict.pop(field)
            row_dict.pop(field)
        computed_tta_column_fields = {"sentiment_overall", "translated_value", "topics"}
        for computed_field in computed_tta_column_fields:
            expected_dict["columns"][0].pop(computed_field)
            row_dict["columns"][0].pop(computed_field)
        expected_dict["columns"][1].update({"value": 100000})
        self.assertDictEqual(row_dict, expected_dict)



from sys import stderr

from caplena.logging.logger import Logger, LoggingLevel


class DefaultLogger(Logger):
    def log(self, msg: str, *, level: LoggingLevel, **extra: str) -> None:
        msg = f"{level.name}[{self.name}]: {msg}"
        extra_str = [f"{tup[0]}={tup[1]}" for tup in extra.items()]
        if len(extra_str) > 0:
            msg += " (" + ", ".join(extra_str) + ")"

        if level.level >= self._logging_level.level:
            print(msg, file=stderr)



from enum import Enum


class ApiBaseUri(Enum):
    LOCAL = "http://localhost:8000/v2"
    PRODUCTION = "https://api.caplena.com/v2"

    @property
    def url(self) -> str:
        return self.value



import copy
from typing import Dict, List, Optional, Tuple, Type, TypeVar

from typing_extensions import Literal

from caplena.helpers import Helpers

T = TypeVar("T", bound="ApiOrdering")
Ordering = List[Tuple[Literal["asc", "desc"], str]]


class ApiOrdering:
    def __init__(self, ordering: Optional[Ordering] = None):
        if ordering is None:
            self._ordering = []
        else:
            self._ordering = ordering

    def to_query_params(self) -> Dict[str, str]:
        stringified_ordering: List[str] = []
        for (direction, name) in self._ordering:
            name = Helpers.build_escaped_filter_str(name)
            stringified_ordering.append(f"{direction}:{name}")
        return {"order_by": ";".join(stringified_ordering)}

    def __str__(self) -> str:
        stringified_ordering: List[str] = []
        for (direction, name) in self._ordering:
            stringified_ordering.append(f"{direction}({name})")
        return "ApiOrdering(" + ", ".join(stringified_ordering) + ")"

    @classmethod
    def asc(cls: Type[T], name: str) -> T:
        return cls(ordering=[("asc", name)])

    @classmethod
    def desc(cls: Type[T], name: str) -> T:
        return cls(ordering=[("desc", name)])

    def __and__(self: T, other: T) -> T:
        new_ordering = copy.deepcopy(self._ordering)
        other_ordering = copy.deepcopy(other._ordering)
        new_ordering.extend(other_ordering)

        return type(self)(ordering=new_ordering)



from copy import deepcopy
from typing import (
    Any,
    Callable,
    ClassVar,
    Dict,
    Generic,
    Iterable,
    List,
    Optional,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
)

from caplena.api import ApiFilter, ApiOrdering
from caplena.api.api_requestor import ApiRequestor
from caplena.configuration import Configuration
from caplena.constants import NOT_SET
from caplena.helpers import Helpers
from caplena.http.http_response import HttpResponse
from caplena.iterator import CaplenaIterator
from caplena.list import CaplenaList

BO = TypeVar("BO", bound="BaseObject[Any]")
BC = TypeVar("BC", bound="BaseController")
T = TypeVar("T")


class BaseController:
    DEFAULT_ALLOWED_CODES: ClassVar[Iterable[int]] = frozenset({200})
    DEFAULT_ALLOWED_POST_CODES: ClassVar[Iterable[int]] = frozenset({201})
    DEFAULT_ALLOWED_DELETE_CODES: ClassVar[Iterable[int]] = frozenset({204})

    @property
    def config(self) -> Configuration:
        return self._config

    @property
    def api(self) -> ApiRequestor:
        return self._config.api_requestor

    def __init__(self, *, config: Configuration):
        self._config = config

    def build(self, resource: Type[BO], obj: Dict[str, Any]) -> BO:
        return resource.build_obj(obj=obj, controller=self, obj_exists=False)

    def get(
        self,
        path: str,
        *,
        allowed_codes: Iterable[int] = DEFAULT_ALLOWED_CODES,
        path_params: Optional[Dict[str, str]] = None,
        query_params: Optional[Dict[str, str]] = None,
        filter: Optional[ApiFilter] = None,
        order_by: Optional[ApiOrdering] = None,
    ) -> HttpResponse:
        response = self._config.api_requestor.get(
            base_uri=self._config.api_base_uri,
            path=path,
            api_key=self._config.api_key,
            api_version=self._config.api_version,
            path_params=path_params,
            query_params=query_params,
            filter=filter,
            order_by=order_by,
        )

        if response.status_code not in allowed_codes:
            raise self._config.api_requestor.build_exc(response)

        return response

    def post(
        self,
        path: str,
        *,
        allowed_codes: Iterable[int] = DEFAULT_ALLOWED_POST_CODES,
        path_params: Optional[Dict[str, str]] = None,
        query_params: Optional[Dict[str, str]] = None,
        json: Optional[Union[Dict[str, Any], List[Any]]] = None,
    ) -> HttpResponse:
        response = self._config.api_requestor.post(
            base_uri=self._config.api_base_uri,
            path=path,
            api_key=self._config.api_key,
            api_version=self._config.api_version,
            path_params=path_params,
            query_params=query_params,
            json=json,
        )

        if response.status_code not in allowed_codes:
            raise self._config.api_requestor.build_exc(response)

        return response

    def put(
        self,
        path: str,
        *,
        allowed_codes: Iterable[int] = DEFAULT_ALLOWED_CODES,
        path_params: Optional[Dict[str, str]] = None,
        query_params: Optional[Dict[str, str]] = None,
        json: Optional[Union[Dict[str, Any], List[Any]]] = None,
    ) -> HttpResponse:
        response = self._config.api_requestor.put(
            base_uri=self._config.api_base_uri,
            path=path,
            api_key=self._config.api_key,
            api_version=self._config.api_version,
            path_params=path_params,
            query_params=query_params,
            json=json,
        )

        if response.status_code not in allowed_codes:
            raise self._config.api_requestor.build_exc(response)

        return response

    def patch(
        self,
        path: str,
        *,
        allowed_codes: Iterable[int] = DEFAULT_ALLOWED_CODES,
        path_params: Optional[Dict[str, str]] = None,
        query_params: Optional[Dict[str, str]] = None,
        json: Optional[Union[Dict[str, Any], List[Any]]] = None,
    ) -> HttpResponse:
        response = self._config.api_requestor.patch(
            base_uri=self._config.api_base_uri,
            path=path,
            api_key=self._config.api_key,
            api_version=self._config.api_version,
            path_params=path_params,
            query_params=query_params,
            json=json,
        )

        if response.status_code not in allowed_codes:
            raise self._config.api_requestor.build_exc(response)

        return response

    def delete(
        self,
        path: str,
        *,
        allowed_codes: Iterable[int] = DEFAULT_ALLOWED_DELETE_CODES,
        path_params: Optional[Dict[str, str]] = None,
        query_params: Optional[Dict[str, str]] = None,
    ) -> HttpResponse:
        response = self._config.api_requestor.delete(
            base_uri=self._config.api_base_uri,
            path=path,
            api_key=self._config.api_key,
            api_version=self._config.api_version,
            path_params=path_params,
            query_params=query_params,
        )

        if response.status_code not in allowed_codes:
            raise self._config.api_requestor.build_exc(response)

        return response

    def build_response(
        self,
        response: HttpResponse,
        *,
        resource: Type[BO],
        metadata: Optional[Dict[str, Any]] = None,
    ) -> BO:
        json = self._retrieve_json_or_raise(response)
        return resource.build_obj(obj=json, controller=self, obj_exists=True, metadata=metadata)

    def build_iterator(
        self,
        *,
        fetcher: Callable[[int], HttpResponse],
        resource: Type[BO],
        limit: Optional[int] = None,
        metadata: Optional[Dict[str, Any]] = None,
    ) -> CaplenaIterator[BO]:
        def results_fetcher(page: int) -> Tuple[List[BO], bool, int]:
            response = fetcher(page)
            json = self._retrieve_json_or_raise(response)

            results = [
                resource.build_obj(res, controller=self, obj_exists=True, metadata=metadata)
                for res in json["results"]
            ]
            return results, json["next_url"] is not None, json["count"]

        return CaplenaIterator(
            results_fetcher=results_fetcher,
            limit=limit,
        )

    def _retrieve_json_or_raise(self, response: HttpResponse) -> Dict[str, Any]:
        json = response.json
        if json is None:
            raise self.api.build_exc(response)
        else:
            return json


class BaseObject(Generic[BC]):
    __fields__: ClassVar[Set[str]] = set()
    __mutable__: ClassVar[Set[str]] = set()

    _attrs: Dict[str, Any]
    _previous: Dict[str, Any]
    _metadata: Dict[str, Any]
    _controller: Optional[BC]

    @property
    def controller(self) -> BC:
        if self._controller is None:
            raise ValueError(
                "You cannot access the non-existing controller for this object. HINT: This object either "
                "does not have a controller attached, or you forgot to manually set the controller after initializing it. "
                "(object.controller = your_controller)"
            )
        return self._controller

    @property
    def is_modified(self) -> bool:
        return self._attrs != self._previous

    def __init__(self, **attrs: Any):
        self._controller = None
        self._metadata = {}
        self._previous = {}
        self._attrs = Helpers.partial_dict(attrs, self.__fields__)

    def dict(self) -> Dict[str, Any]:
        if self.is_modified:
            raise ValueError(
                "Cannot call `.dict()` on a modified object. HINT: Please call `.save()` to "
                "propagate your updates to our API servers."
            )

        resource: Dict[str, Any] = {}
        for field in self.__fields__:
            attr = self._attrs[field]
            resource[field] = self._rec_dict(attr)

        return resource

    def modified_dict(self) -> Any:
        resource: Dict[str, Any] = {}
        for field in self.__fields__:
            if field not in self._previous:
                resource[field] = self._rec_modified_dict(
                    previous=NOT_SET, next=self._attrs[field], field=field
                )
            elif self._previous[field] != self._attrs[field]:
                resource[field] = self._rec_modified_dict(
                    previous=self._previous[field], next=self._attrs[field], field=field
                )

        return resource if resource != {} else NOT_SET

    def _refresh_from(self, *, attrs: Dict[str, Any]) -> None:
        self._attrs = Helpers.partial_dict(attrs, self.__fields__)
        self._previous = deepcopy(self._attrs)

    def _prepare(
        self,
        *,
        controller: Optional[BC],
        obj_exists: bool = False,
    ) -> None:
        self._controller = controller
        if obj_exists:
            self._previous = deepcopy(self._attrs)

        for field in self.__fields__:
            self._rec_prepare(self._attrs[field], controller=controller, obj_exists=obj_exists)

    def _rec_dict(self, attr: Any) -> Any:
        if isinstance(attr, BaseObject):
            return attr.dict()
        elif isinstance(attr, CaplenaList):
            return [self._rec_dict(i) for i in attr]  # pyright: reportUnknownVariableType=false
        else:
            return attr

    def _rec_modified_dict(self, *, previous: Any, next: Any, field: str) -> Any:
        if isinstance(next, BaseObject):
            return next.modified_dict()
        elif isinstance(next, CaplenaList):
            if len(previous) != len(next):  # pyright: reportUnknownArgumentType=false
                raise ValueError(
                    "Failed computing the version difference, as the previous and new lists have a different length."
                )

            modified_list = [
                self._rec_modified_dict(previous=p, next=n, field=f"{field}.{idx}")
                for idx, (p, n) in enumerate(zip(previous, next))
            ]
            return [item for item in modified_list if item != NOT_SET]

        else:
            return next

    def _rec_prepare(self, attr: Any, *, controller: Optional[BC], obj_exists: bool) -> None:
        if isinstance(attr, BaseObject):
            attr._prepare(
                controller=controller, obj_exists=obj_exists
            )  # pyright: reportUnknownMemberType=false
        elif isinstance(attr, CaplenaList):
            for i in attr:
                self._rec_prepare(i, controller=controller, obj_exists=obj_exists)

    def __str__(self) -> str:
        return f"{self.__class__.__name__}()"

    def __setattr__(self, name: str, value: Any) -> None:
        # TODO: interface for patching can still be improved, currently customer has to
        # manually instantiate the class instance, e.g. when adding new topics.
        if name in self.__fields__ and name in self.__mutable__:
            self._attrs[name] = value
        elif name in self.__fields__:
            raise AttributeError(
                f"{name}. HINT: You cannot modify this attribute, as it is immutable."
            )
        else:
            super().__setattr__(name, value)

    def __getattr__(self, name: str) -> Any:
        if name in self.__fields__:
            return self._attrs[name]
        else:
            return super().__getattribute__(name)

    def __delattr__(self, name: str) -> None:
        raise ValueError(f"{name}. HINT: You cannot delete any attributes.")

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, BaseObject):
            return False

        if self.__fields__ != other.__fields__:
            return False

        for field in self.__fields__:
            if self._attrs[field] != other._attrs[field]:
                return False

        return True

    @classmethod
    def build_obj(
        cls: Type[BO],
        obj: Dict[str, Any],
        *,
        controller: Optional[BC],
        obj_exists: bool = False,
        metadata: Optional[Dict[str, Any]] = None,
    ) -> BO:
        instance = cls.parse_obj(obj)
        instance._prepare(controller=controller, obj_exists=obj_exists)
        instance._metadata = metadata if metadata else {}

        return instance

    @classmethod
    def parse_obj(cls: Type[BO], obj: Dict[str, Any]) -> BO:
        return cls(**obj)


class BaseResource(BaseObject[BC]):
    @property
    def id(self) -> str:
        return self._id

    def __init__(self, id: str, **attrs: Any):
        self._id = id
        super().__init__(**attrs)

    def __str__(self) -> str:
        return f"{self.__class__.__name__}(id={self._id})"

    def dict(self) -> Dict[str, Any]:
        resource = super().dict()
        resource["id"] = self._id
        return resource



from typing import Any, Dict


class _NOT_SET:
    def __str__(self) -> str:
        return "<NOT_SET>"

    def __copy__(self) -> Any:
        return NOT_SET

    def __deepcopy__(self, memodict: Dict[str, Any] = {}) -> Any:
        return NOT_SET


# note: we use NOT_SET in order to differentiate between `null` and
# `undefined` in JSON. this is especially relevant for update requests
# where not all properties need to be present.
NOT_SET: Any = _NOT_SET()

# Pagination limits
LIST_PAGINATION_LIMIT = 30



from enum import Enum


class ApiVersion(Enum):
    DEFAULT = 0
    VER_2022_01_01 = 1
    VER_2022_06_20 = 2

    @property
    def version(self) -> str:
        if self.name != ApiVersion.DEFAULT.name:
            return self.name.replace("VER_", "").replace("_", "-")
        else:
            raise ValueError(f"Cannot convert `{self.name}` to a valid version string.")



from datetime import datetime
from json import JSONEncoder
from typing import Any

from caplena.helpers import Helpers


class JsonDateEncoder(JSONEncoder):
    def default(self, o: Any) -> Any:
        if isinstance(o, datetime):
            return Helpers.to_rfc3339_datetime(o)
        return JSONEncoder.default(self, o)
